@@ -3,7 +3,7 @@ require LWP::UserAgent;
 
 use vars qw(@ISA $VERSION);
 @ISA = qw(LWP::UserAgent);
-$VERSION = '1.00';
+$VERSION = '1.08_01';
 
 require HTTP::Request;
 require HTTP::Response;
@@ -11,6 +11,7 @@ require HTTP::Response;
 use HTTP::Status ();
 use strict;
 use Net::DNS;
+use LWP::Debug ();
 
 sub new {
     my $class = shift;
@@ -19,6 +20,7 @@ sub new {
     my $blocked_hosts     = delete $opts{blocked_hosts}     || [];
     my $whitelisted_hosts = delete $opts{whitelisted_hosts} || [];
     my $resolver          = delete $opts{resolver};
+    my $paranoid_proxy    = delete $opts{paranoid_proxy};
     $opts{timeout}      ||= 15;
 
     my $self = LWP::UserAgent->new( %opts );
@@ -26,6 +28,7 @@ sub new {
     $self->{'blocked_hosts'}     = $blocked_hosts;
     $self->{'whitelisted_hosts'} = $whitelisted_hosts;
     $self->{'resolver'}          = $resolver;
+    $self->{'paranoid_proxy'}    = $paranoid_proxy;
 
     $self = bless $self, $class;
     return $self;
@@ -47,7 +50,7 @@ sub _resolve {
     $depth ||= 0;
 
     die "CNAME recursion depth limit exceeded.\n" if $depth > 10;
-    die "Suspicious results from DNS lookup" if $self->_bad_host($host);
+    die "DNS lookup resulted in bad host." if $self->_bad_host($host);
 
     # return the IP address if it looks like one and wasn't marked bad
     return ($host) if $host =~ /^\d+\.\d+\.\d+\.\d+$/;
@@ -55,10 +58,14 @@ sub _resolve {
     my $sock = $res->bgsend($host)
         or die "No sock from bgsend";
 
-    my $rin = '';
-    vec($rin, fileno($sock), 1) = 1;
-    my $nf = select($rin, undef, undef, $self->_time_remain($request));
-    die "DNS lookup timeout" unless $nf;
+    # wait for the socket to become readable, unless this is from our test
+    # mock resolver.
+    unless ($sock && $sock eq "MOCK") {
+        my $rin = '';
+        vec($rin, fileno($sock), 1) = 1;
+        my $nf = select($rin, undef, undef, $self->_time_remain($request));
+        die "DNS lookup timeout" unless $nf;
+    }
 
     my $packet = $res->bgread($sock)
         or die "DNS bgread failure";
@@ -69,7 +76,8 @@ sub _resolve {
     foreach my $rr ($packet->answer) {
         if ($rr->type eq "A") {
             die "Suspicious DNS results from A record\n" if $self->_bad_host($rr->address);
-            push @addr, $rr->address;
+            # untaints the address:
+            push @addr, join(".", ($rr->address =~ /^(\d+)\.(\d+)\.(\d+)\.(\d+)$/));
         } elsif ($rr->type eq "CNAME") {
             # will be checked for validity in the recursion path
             $cname = $rr->cname;
@@ -86,7 +94,7 @@ sub _host_list_match {
     my $list_name = shift;
     my $host = shift;
 
-    foreach my $rule (@{ $self->{$list_name} }) {
+    foreach my $rule (@{ $self->{$list_name} || [] }) {
         if (ref $rule eq "CODE") {
             return 1 if $rule->($host);
         } elsif (ref $rule) {
@@ -103,21 +111,82 @@ sub _bad_host {
     my $host = lc(shift);
 
     return 0 if $self->_host_list_match("whitelisted_hosts", $host);
+    return 1 if $self->_host_list_match("blocked_hosts", $host);
+    return 1 if
+        $host =~ /^localhost$/i ||    # localhost is bad.  even though it'd be stopped in
+                                      #    a later call to _bad_host with the IP address
+        $host =~ /\s/i;               # any whitespace is questionable
 
-    # don't let people connect to private, internal, or multicast IP addresses
-    if ($host =~ /^(?:10\.|127\.|192\.168\.)/ ||
-        ($host =~ /^172\.(\d+)/ && ($1 >= 16 && $1 <= 31)) ||
-        ($host =~ /^2(\d+)/ && ($1 >= 24 && $1 <= 54)) ||
-        $host =~ /^localhost$/i ||
-        $host =~ /\b0\d+\./ || # octal IPs: 0177.0.0.1 == 127.0.0.1
-        $host =~ /\s/ ||       # whitespace
-        $host =~ /^\d+$/       # purely decimal/octal IP addresses http://167838209/ == 10.1.2.1
-        ) {
+    # Let's assume it's an IP address now, and get it into 32 bits.
+    # Uf at any time something doesn't look like a number, then it's
+    # probably a hostname and we've already either whitelisted or
+    # blacklisted those, so we'll just say it's okay and it'll come
+    # back here later when the resolver finds an IP address.
+    my @parts = split(/\./, $host);
+    return 0 if @parts > 4;
+
+    # un-octal/un-hex the parts, or return if there's a non-numeric part
+    my $overflow_flag = 0;
+    foreach (@parts) {
+        return 0 unless /^\d+$/ || /^0x[a-f\d]+$/;
+        local $SIG{__WARN__} = sub { $overflow_flag = 1; };
+        $_ = oct($_) if /^0/;
+    }
+
+    # a purely numeric address shouldn't overflow.
+    return 1 if $overflow_flag;
+
+    my $addr;  # network order packed IP address
+
+    if (@parts == 1) {
+        # a - 32 bits
+        return 1 if
+            $parts[0] > 0xffffffff;
+        $addr = pack("N", $parts[0]);
+    } elsif (@parts == 2) {
+        # a.b - 8.24 bits
+        return 1 if
+            $parts[0] > 0xff ||
+            $parts[1] > 0xffffff;
+        $addr = pack("N", $parts[0] << 24 | $parts[1]);
+    } elsif (@parts == 3) {
+        # a.b.c - 8.8.16 bits
+        return 1 if
+            $parts[0] > 0xff ||
+            $parts[1] > 0xff ||
+            $parts[2] > 0xffff;
+        $addr = pack("N", $parts[0] << 24 | $parts[1] << 16 | $parts[2]);
+    } elsif (@parts == 4) {
+        # a.b.c.d - 8.8.8.8 bits
+        return 1 if
+            $parts[0] > 0xff ||
+            $parts[1] > 0xff ||
+            $parts[2] > 0xff ||
+            $parts[3] > 0xff;
+        $addr = pack("N", $parts[0] << 24 | $parts[1] << 16 | $parts[2] << 8 | $parts[3]);
+    } else {
         return 1;
     }
 
-    return 1 if $self->_host_list_match("blocked_hosts", $host);
+    my $haddr = unpack("N", $addr); # host order IP address
+    return 1 if
+        ($haddr & 0xFF000000) == 0x00000000 || # 0.0.0.0/8
+        ($haddr & 0xFF000000) == 0x0A000000 || # 10.0.0.0/8
+        ($haddr & 0xFF000000) == 0x7F000000 || # 127.0.0.0/8
+        ($haddr & 0xFFF00000) == 0xAC100000 || # 172.16.0.0/12
+        ($haddr & 0xFFFF0000) == 0xA9FE0000 || # 169.254.0.0/16
+        ($haddr & 0xFFFF0000) == 0xC0A80000 || # 192.168.0.0/16
+        ($haddr & 0xFFFFFF00) == 0xC0000200 || # 192.0.2.0/24  "TEST-NET" docs/example code
+        ($haddr & 0xFFFFFF00) == 0xC0586300 || # 192.88.99.0/24 6to4 relay anycast addresses
+         $haddr               == 0xFFFFFFFF || # 255.255.255.255
+        ($haddr & 0xF0000000) == 0xE0000000;  # multicast addresses
 
+    # as final IP address check, pass in the canonical a.b.c.d decimal form
+    # to the blacklisted host check to see if matches as bad there.
+    my $can_ip = join(".", map { ord } split //, $addr);
+    return 1 if $self->_host_list_match("blocked_hosts", $can_ip);
+
+    # looks like an okay IP address
     return 0;
 }
 
@@ -141,6 +210,12 @@ sub request {
         return $err_res;
     }
 
+    if (my $pp = $self->{paranoid_proxy}) {
+        $req->uri("$pp?url="   . eurl($req->uri) .
+                  "&timeout="  . ($self->{timeout}  + 0) .
+                  "&max_size=" . ($self->{max_size} + 0));
+    }
+
     return $self->SUPER::request($req, $arg, $size, $previous);
 }
 
@@ -239,11 +314,12 @@ EOT
             $response = $protocol->request($request, $proxy,
                                            $arg, $size, $timeout);
         };
-        if ($@) {
-            $@ =~ s/ at .* line \d+.*//s;  # remove file/line number
+        my $error = $@ || $response->header( 'x-died' );
+        if ($error) {
+            $error =~ s/ at .* line \d+.*//s; # remove file/line number
             $response = _new_response($request,
                                       &HTTP::Status::RC_INTERNAL_SERVER_ERROR,
-                                      $@);
+                                      $error);
         }
     }
     else {
@@ -255,6 +331,7 @@ EOT
     $response->request($request);  # record request for reference
     $cookie_jar->extract_cookies($response) if $cookie_jar;
     $response->header("Client-Date" => HTTP::Date::time2str(time));
+    $self->run_handlers("response_done", $response) if $self->can('run_handlers');
     return $response;
 }
 
@@ -267,7 +344,7 @@ sub blocked_hosts
         $self->{'blocked_hosts'} = \@hosts;
         return;
     }
-    return @{ $self->{'blocked_hosts'} };
+    return @{ $self->{'blocked_hosts'} || [] };
 }
 
 # whitelisted hostnames, compiled patterns, or subrefs
@@ -279,7 +356,7 @@ sub whitelisted_hosts
         $self->{'whitelisted_hosts'} = \@hosts;
         return;
     }
-    return @{ $self->{'whitelisted_hosts'} };
+    return @{ $self->{'whitelisted_hosts'} || [] };
 }
 
 # get/set Net::DNS resolver object
@@ -304,7 +381,7 @@ sub _need_proxy
 
     my $scheme = $url->scheme || return;
     if (my $proxy = $self->{'proxy'}{$scheme}) {
-        if (@{ $self->{'no_proxy'} }) {
+        if ($self->{'no_proxy'} && @{ $self->{'no_proxy'} }) {
             if (my $host = eval { $url->host }) {
                 for my $domain (@{ $self->{'no_proxy'} }) {
                     if ($host =~ /\Q$domain\E$/) {
@@ -354,6 +431,13 @@ sub _new_response {
     return $response;
 }
 
+sub eurl {
+    my $a = $_[0];
+    $a =~ s/([^a-zA-Z0-9_\,\-.\/\\\: ])/uc sprintf("%%%02x",ord($1))/eg;
+    $a =~ tr/ /+/;
+    return $a;
+}
+
 1;
 
 __END__
@@ -426,10 +510,67 @@ tarpitting your process (the timeout parameter is changed to be a global
 timeout over the entire process), and all combinations of redirects and
 DNS tricks to otherwise tarpit and/or connect to internal resources.
 
+=head1 CONSTRUCTOR
+
+=over 4
+
+=item C<new>
+
+my $ua = LWPx::ParanoidAgent->new([ %opts ]);
+
+In addition to any constructor options from L<LWP::UserAgent>, you may
+also set C<blocked_hosts> (to an arrayref), C<whitelisted_hosts> (also
+an arrayref), and C<resolver>, a Net::DNS::Resolver object.
+
+=back
+
+=head1 METHODS
+
+=over 4
+
+=item $csr->B<resolver>($net_dns_resolver)
+
+=item $csr->B<resolver>
+
+Get/set the L<Net::DNS::Resolver> object used to lookup hostnames.
+
+=item $csr->B<blocked_hosts>(@host_list)
+
+=item $csr->B<blocked_hosts>
+
+Get/set the the list of blocked hosts.  The items in @host_list may be
+compiled regular expressions (with qr//), code blocks, or scalar
+literals.  In any case, the thing that is match, passed in, or
+compared (respectively), is all of the given hostname, given IP
+address, and IP address in canonical a.b.c.d decimal notation.  So if
+you want to block "1.2.3.4" and the user entered it in a mix of
+network/host form in a mix of decimal/octal/hex, you need only block
+"1.2.3.4" and not worry about the details.
+
+=item $csr->B<whitelisted_hosts>(@host_list)
+
+=item $csr->B<whitelisted_hosts>
+
+Like blocked hosts, but matching the hosts/IPs that bypass blocking
+checks.  The only difference is the IP address isn't canonicalized
+before being whitelisted-matched, mostly because it doesn't make sense
+for somebody to enter in a good address in a subversive way.
+
+=back
+
 =head1 SEE ALSO
 
 See L<LWP::UserAgent> to see how to use this class.
 
+http://contributing.appspot.com/lwpx-paranoidagent
+http://brad.livejournal.com/2409049.html
+https://github.com/collectiveintel/LWPx-ParanoidAgent
+http://search.cpan.org/dist/LWPx-ParanoidAgent
+
+=head1 ISSUES
+
+Report issues: https://github.com/collectiveintel/LWPx-ParanoidAgent/issues
+
 =head1 WARRANTY
 
 This module is supplied "as-is" and comes with no warranty, expressed
@@ -443,7 +584,8 @@ Maybe.  See the warranty above.
 
 =head1 COPYRIGHT
 
-Copyright 2005 Brad Fitzpatrick
+ Copyright 2005 Brad Fitzpatrick
+ Copyright 2013 Wes Young (wesyoung.me)
 
 Lot of code from the the base class, copyright 1995-2004 Gisle Aas.
 
